/*
 * \file soss.c
 *
 * Main code of SOSS, Simple OSS Sound.
 *
 * \author Pontus Lidman <pontus@lysator.liu.se>
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License. See the file "COPYING" in this directory for
 * details.
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>

#include <linux/soundcard.h>
#include <fcntl.h>

#include "soss.h"

/* Fixed-point precision of frequency values */
#define FREQ_PREC 256UL

/* Fixed-point precision of per-sample volume ramps. This value is
   hard-coded in the asm mixer. */
#define RAMP_PREC 1024

/* not negotiable; asm code gets high bits of multiplication
   result where C code uses this one. */
#define VOLUME_PREC 65536 

static stereo_t soundbuffer[BUF_SAMPLES] __attribute__((aligned(16)));
stereo_t *backBuffer = soundbuffer;

channel_t channels[NUM_CHAN];
chanext_t chanexts[NUM_CHAN];

void play_note(instrument_t *instr, uint32_t freq, int chan, int keydown)
{
    channel_t *ch = &channels[chan];
    ch->step = (FREQ_PREC*freq)/SAMPLE_RATE;
    ch->ext->keydown = keydown;
    if (keydown) {
	ch->ext->instr = instr;
	ch->samples = instr->samples;
	ch->nsamples = instr->nsamples;
	ch->pos = 0;
	ch->posfr = 0;
	ch->ext->envpos = 0;
    } else {
	ch->ext->releaselevel = ch->ext->adsrvolume;
	ch->ext->releasetime = ch->ext->envpos;
    }
}

int32_t adsr(channel_t *ch, instrument_t *instr, uint32_t t, int keydown)
{
    int32_t r;

    if (keydown) {
	if (t < instr->attack) {
	    r = t * (VOLUME_PREC-1) / instr->attack;
	    return r;
	}
	t -= instr->attack;
	if (t < instr->decay) {
	    r = (instr->decay - t) * ((VOLUME_PREC-1)-instr->sustain) / instr->decay + instr->sustain;
	    return r;
	}
	return instr->sustain;
    } else {
	t -= ch->ext->releasetime;
	if (t < instr->release) {
	    r = ch->ext->releaselevel - (t * ch->ext->releaselevel / instr->release);
	    return r;
	} else {
	    return 0;
	}
    }
}

void pan(channel_t *ch, int32_t t, int32_t *left, int32_t *right)
{
    int32_t p;
    instrument_t *instr = ch->ext->instr;

    if (t < instr->pan_duration) {
	int32_t delta;
	delta = instr->pan_end - instr->pan_start;
	p = instr->pan_start + (t * delta) / instr->pan_duration;
    } else {
	p = instr->pan_end;
    }
    if (p > 0) {
	*left = (*left * (32768-p)) / 32768;
    } else {
	*right = (*right * (32768+p)) / 32768;
    }
}

void envelope(void)
{
    int i;
    int32_t vol1, vol2, vol1l, vol1r, vol2l, vol2r;
    for (i=0; i<NUM_CHAN; i++) {
	channel_t *ch = &channels[i];
	if (!ch->ext->instr)
	    continue;

	vol1 = adsr(ch, ch->ext->instr, ch->ext->envpos, ch->ext->keydown);
	ch->ext->adsrvolume = vol1;

	vol2 = adsr(ch, ch->ext->instr, ch->ext->envpos+1, ch->ext->keydown);

	vol1l = vol1;
	vol1r = vol1;
	pan(ch, ch->ext->envpos, &vol1l, &vol1r);
	vol2l = vol2;
	vol2r = vol2;
	pan(ch, ch->ext->envpos+1, &vol2l, &vol2r);

	ch->volume_left = (vol1l*ch->ext->master) / VOLUME_PREC;
	ch->volume_right = (vol1r*ch->ext->master) / VOLUME_PREC;
	ch->vramp_l = (vol2l - vol1l)*RAMP_PREC/BUF_SAMPLES;
	ch->vramp_r = (vol2r - vol1r)*RAMP_PREC/BUF_SAMPLES;
	ch->vramp_l = (ch->vramp_l * ch->ext->master) / VOLUME_PREC;
	ch->vramp_r = (ch->vramp_r * ch->ext->master) / VOLUME_PREC;
	ch->ext->envpos++;
    }
}


#define clamp16(si) ((si) > 32767 ? 32767 : (si) < -32768 ? -32768 : (si))

void c_mixer(void)
{
    int32_t left, right;
    int i, j;
    stereo_t outsmp;
    mono_t wave;
    for (j=0; j<NUM_CHAN; j++) {
	channels[j].ramptmp_l = 0;
	channels[j].ramptmp_r = 0;
    }
    for (i=0; i<BUF_SAMPLES; i++) {
	left = 0;
	right = 0;
	for (j=0; j<NUM_CHAN; j++) {
	    channel_t *ch = &channels[j];
	    if (!ch->samples)
		continue;
	    wave = ch->samples[ch->pos/2];
	    ch->posfr += ch->step;
	    ch->pos += 2*(ch->posfr / FREQ_PREC);
	    ch->posfr &= (FREQ_PREC-1);
	    if (ch->pos >= 2*ch->ext->instr->nsamples)
		ch->pos -= 2*ch->ext->instr->nsamples;

	    left += (ch->volume_left + (ch->ramptmp_l/RAMP_PREC))*wave / VOLUME_PREC;
	    right += (ch->volume_right + (ch->ramptmp_r/RAMP_PREC))*wave / VOLUME_PREC;
	    ch->ramptmp_l += ch->vramp_l;
	    ch->ramptmp_r += ch->vramp_r;
	}
	outsmp.left = clamp16(left);
	outsmp.right = clamp16(right);
	backBuffer[i] = outsmp;
    }
}

// static uint64_t lastoutput = 0;

void soundoutput(int fd, int dumpfd)
{
#if 0
    desired = lastoutput + 16600;
    while (1) {
	gettimeofday(&now, 0);
	now64 = (now.tv_sec * 1000000) + now.tv_usec;
	if (now64 >= desired)
	    break;
    }
    lastoutput = now64;
#endif
    write(fd, backBuffer, sizeof(stereo_t)*BUF_SAMPLES);
    if (dumpfd >= 0) {
	write(dumpfd, backBuffer, sizeof(stereo_t)*BUF_SAMPLES);
    }
}

int oss_open(void)
{
    /*
     * PS2 SPU native format is AF_S16_LE, 2 channels, 48kHz.
     * Open OSS device with those parameters or bust.
     */
    int audio_fd = -1;
    int format = AFMT_S16_LE;
    int channels = 2;
    int speed = SAMPLE_RATE;
    int frag_size;

    audio_fd = open("/dev/dsp", O_WRONLY);

    if (audio_fd < 0) {
	perror("open audio device");
	exit(1);
    }
    
    if (0 > ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format))
	exit(1);

    if (format != AFMT_S16_LE)
	exit(1);

    if (0 > ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &channels))
	exit(1);

    if (channels != 2)
	exit(1);

    if (0 > ioctl(audio_fd, SNDCTL_DSP_SPEED, &speed))
	exit(1);

    if (speed != SAMPLE_RATE)
	exit(1);

    if (0 > ioctl(audio_fd, SNDCTL_DSP_GETBLKSIZE, &frag_size))
	exit(1);
    return audio_fd;
}

#define TEST_WAVE_LEN 512

mono_t sinwave[TEST_WAVE_LEN];
mono_t sawwave[TEST_WAVE_LEN];
mono_t sqwave[TEST_WAVE_LEN];
mono_t noisewave[TEST_WAVE_LEN];
mono_t triwave[TEST_WAVE_LEN];

instrument_t sininstr = {
    sinwave,
    TEST_WAVE_LEN,
    4, 15, (VOLUME_PREC-1)/2, 30,
    0, 0, 0
};

instrument_t sin2instr = {
    sinwave,
    TEST_WAVE_LEN,
    4, 15, (VOLUME_PREC-1)/2, 30,
    -32767, 32767, 20
};

instrument_t sin3instr = {
    sinwave,
    TEST_WAVE_LEN,
    4, 15, (VOLUME_PREC-1)/2, 30,
    32767, -32767, 20
};

instrument_t sawinstr = {
    sawwave,
    TEST_WAVE_LEN,
    0, 120, 0, 0,
    0, 0, 0
};

instrument_t sqinstr = {
    sqwave,
    TEST_WAVE_LEN,
    4, 60, 0, 0,
    0, 0, 0
};

instrument_t noiseinstr = {
    noisewave,
    TEST_WAVE_LEN,
    4, 60, 0, 0,
    0, 0, 0
};

instrument_t triinstr = {
    triwave, TEST_WAVE_LEN,
    1, 60, 0, 0,
    0, 0, 0
};

/*
 * Create waveforms for the sample instruments. Note that the sawtooth
 * wave doesn't sound too good, and the noise waveform is even worse
 * (it should be bandwidth-limited).
 */
void initsamples(void)
{
    int i;
    for (i=0; i<TEST_WAVE_LEN; i++) {
	sinwave[i] = 32767.0*sin(i*2*M_PI/TEST_WAVE_LEN);
	if (i<50) {
	    sqwave[i] = 32767;
	} else {
	    sqwave[i] = -32768;
	}
	sawwave[i] = (i-(TEST_WAVE_LEN/2))*(32767/(TEST_WAVE_LEN/2));
	noisewave[i] = lrand48();
	triwave[i] = abs(2*i-TEST_WAVE_LEN) * (32767/TEST_WAVE_LEN);
    }
}

/*
 * Initialize the channel structs by zeroing them, initializing the
 * master volume, and attaching the extension structures.
 */
void channel_init(void)
{
    int i;
    for (i=0; i<NUM_CHAN; i++) {
	memset(&channels[i], 0, sizeof(channels[i]));
	memset(&chanexts[i], 0, sizeof(chanexts[i]));
	channels[i].ext = &chanexts[i];
	channels[i].ext->master = VOLUME_PREC/4;
    }
}
